#!/usr/bin/python
# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of the Willow Garage nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Author: Wim Meeussen




from __future__ import with_statement

import roslib; roslib.load_manifest('application_manager')
import rospy
import os
import threading
import actionlib
from actionlib_msgs.msg import GoalStatus
import yaml
import roslauncher
from application_msgs.msg import *




class Application():
    def __init__(self):
        self.state = 'Stopped' # state of the application (Running, Stopped, Preempted)
        self.client_goals = [] # list of clients that are connected to this application
        

    def configure(self, conf):
        self.lock = threading.Lock()  # lock for interacting with this application

        # check required keys
        keys = ['package', 'name', 'description', 'launch_file']
        for k in keys:
            if not k in conf:
                rospy.logerr('Application in package %s does not specify option "%s"'%(conf['package'], k))
                return False

        self.name = '/'.join([conf['package'], conf['name']])
        self.description = conf['description']
        try:
            with open(os.path.join(roslib.packages.get_pkg_dir(conf['package']), conf['launch_file'])) as f:
                self.launch_file = f.read()
        except:
            rospy.logerr('Application %s specified invalid package or launch file'%conf['name'])
            return False

        # optional action interface
        self.action_ns = None
        if 'action_ns' in conf and 'preempt_timeout' in conf:
            self.action_ns = conf['action_ns']
            self.preempt_timeout = rospy.Duration(conf['preempt_timeout'])
            self.launch_timeout = rospy.Duration(30.0)
            self.action_client = actionlib.SimpleActionClient(self.action_ns, ApplicationAction)

        # optional daemon parameter
        self.daemonize = False
        if 'daemonize' in conf:
            self.daemonize = conf['daemonize']

        return True



    def start(self):
        with self.lock:
            rospy.loginfo('Starting application %s'%self.name)
            self.launcher = roslauncher.ScriptRoslaunch(self.launch_file)
            self.launcher.start()

            # start action client if application supports this
            if self.action_ns:
                rospy.loginfo('Waiting for action client to come up in namespace %s for application %s'%(self.action_ns, self.name))
                if not self.action_client.wait_for_server(self.launch_timeout):
                    rospy.logerr('Action client server for application %s did not connect in time.'%self.name)                    
                    self.launcher.shutdown()
                    self.launcher = None
                    return False
                self.action_client.send_goal(ApplicationGoal())
                start_time = rospy.Time.now()
                while self.action_client.get_state() != GoalStatus.ACTIVE:
                    rospy.sleep(0.1)
                    if rospy.Time.now() > start_time + self.launch_timeout:
                        rospy.logerr('Action client server for application %s did not become active in time.'%self.name)
                        self.launcher.shutdown()
                        self.launcher = None
                        return False

            self.state = 'Running'
            rospy.loginfo('Started application %s'%self.name)
            return True


    def stop(self, force):
        self.lock.acquire() # this lock gets released from other thread

        # notify all clients that are still connected
        for g in self.client_goals:
            g.set_canceled()
        self.client_goals = []

        # shut down application in separate thread, so we don't block the manager
        self.application_stopper = ApplicationStopper(self, force)
        self.application_stopper.start()
        return True


class ApplicationStopper(threading.Thread):
    def __init__(self, application, force):
        threading.Thread.__init__(self)
        self.application = application
        self.force = force

    def run(self):
        rospy.loginfo('Application stopper thread created')
        # preempt application if this is supported
        if self.application.action_ns and not self.force:
            rospy.loginfo("Sending preempt signal to application %s"%self.application.name)
            self.application.state = 'Preempted'
            self.application.action_client.cancel_goal()
            self.application.action_client.wait_for_result(self.application.preempt_timeout)

        # kill launch file
        self.application.launcher.shutdown()
        self.application.launcher = None
        self.application.state = 'Stopped'
        rospy.loginfo('Stopped application %s'%self.application.name)
        self.application.lock.release() # this lock got acquired in other thread



class ApplicationManager():
    def __init__(self):
        self.lock = threading.RLock()
        self.trigger_start_app = None
        self.trigger_stop_app = None
        self.trigger_start_result = False
        self.trigger_stop_result = False

        self.applications = self.get_available_applications()
        rospy.loginfo("Found the following applications: %s"%str(self.applications.keys()))
        
        # interface of application manager
        self.run_interface = actionlib.ActionServer('application_manager/run_application', RunApplicationAction, self.run_cb, self.cancel_cb)
        self.kill_interface = actionlib.SimpleActionServer('application_manager/kill_application', KillApplicationAction, self.kill_cb)
        self.stop_interface = actionlib.SimpleActionServer('application_manager/stop_application', StopApplicationAction, self.stop_cb)
        self.list_interface = actionlib.SimpleActionServer('application_manager/list_applications', ListApplicationsAction, self.list_cb)


    def get_available_applications(self):
        with self.lock:
            path = path = roslib.packages.get_pkg_dir('application_manager')
            apps_yaml = yaml.load(open(path+'/applications.yaml'))
            available_applications = {}
            for app in apps_yaml:
                application = Application()
                if not 'app' in app:
                    rospy.logerr('Invalid entry in applications.yaml file: %s'%str(app))
                elif application.configure(app['app']):
                    available_applications[application.name] = application
                    rospy.loginfo("Available application: %s"%application.name)                    
            return available_applications


    def start_application(self, application_name):
        self.trigger_start_app = application_name
        while self.trigger_start_app:
            rospy.sleep(0.1)
        return self.trigger_start_result


    def stop_application(self, application_name, force=False):
        self.force_stop = force
        self.trigger_stop_app = application_name
        while self.trigger_stop_app:
            rospy.sleep(0.1)
        return self.trigger_stop_result
            

    def trigger_start(self, application_name):
        # check if application exists
        if not application_name in self.applications:
            rospy.logerr('Cannot start application %s, no application exists with that name'%application_name)
            return False
        app = self.applications[application_name]

        # check if application is already running
        if app.state == 'Running':
            rospy.loginfo('Applicaiton %s already running, not starting it again'%application_name)
            return True

        # start application
        return app.start()



    def trigger_stop(self, application_name, force):
        # check if application exists
        if not application_name in self.applications:
            rospy.logerr('Cannot stop application %s, no application exists with that name'%application_name)
            return False
        app = self.applications[application_name]

        # check if application is running
        if app.state != 'Running':
            rospy.loginfo("Applicaiton %s is not running, so don't need to stop it"%application_name)
            return True

        # stop application
        return app.stop(force)


    def run_cb(self, goal):
        with self.lock:
            application_name = goal.get_goal().application_name
            if self.start_application(application_name):
                goal.set_accepted()
                self.applications[application_name].client_goals.append(goal)
                rospy.loginfo('Client connected to application %s'%application_name)
            else:
                goal.set_rejected()
            

    def cancel_cb(self, goal):
        with self.lock:
            # cancel goal
            goal.set_canceled()

            # remove goal from app list
            application = None
            for name, app in self.applications.iteritems():
                if goal in app.client_goals:
                    app.client_goals.remove(goal)
                    application = app
            if not application:
                rospy.loginfo('Canceling goal that does not match up with an application any more')
                return
            
            # stop app is no more clients are connected
            if len(application.client_goals) == 0 and not application.daemonize:
                self.stop_application(application.name)
            else:
                rospy.loginfo('Client is disconnected, but keeping application %s running'%application.name)

                
    def stop_cb(self, goal):
        with self.lock:
            if self.stop_application(goal.application_name):
                self.stop_interface.set_succeeded(KillApplicationResult())
            else:
                self.stop_interface.set_aborted(KillApplicationResult())


    def kill_cb(self, goal):
        with self.lock:
            if self.stop_application(goal.application_name, force=True):
                self.kill_interface.set_succeeded(KillApplicationResult())
            else:
                self.kill_interface.set_aborted(KillApplicationResult())


    def list_cb(self, goal):
        with self.lock:
            res = ListApplicationsResult()
            for name, app in self.applications.iteritems():
                res.application_name.append(name)
                res.application_status.append(app.state)
            self.list_interface.set_succeeded(res)


        


def main():
    rospy.init_node('application_manager_node')
    am = ApplicationManager()
    rospy.loginfo("Application manager running")

    while not rospy.is_shutdown(): 
       rospy.sleep(0.1)
       if am.trigger_start_app:
            am.trigger_start_result = am.trigger_start(am.trigger_start_app)
            am.trigger_start_app = None

       if am.trigger_stop_app:
           am.trigger_stop_result = am.trigger_stop(am.trigger_stop_app, am.force_stop)
           am.trigger_stop_app = None




if __name__ == '__main__':
    main()

